@@ -29,6 +29,7 @@ wage as a function of various features such as experience, age, or education.
 
 """
 
+# %%
 import numpy as np
 import scipy as sp
 import pandas as pd
@@ -45,12 +46,12 @@ import seaborn as sns
 
 from sklearn.datasets import fetch_openml
 
-survey = fetch_openml(data_id=534, as_frame=True)
+survey = fetch_openml(data_id=534, as_frame=True, parser="pandas")
 
 # %%
 # Then, we identify features `X` and targets `y`: the column WAGE is our
 # target variable (i.e., the variable which we want to predict).
-#
+
 X = survey.data[survey.feature_names]
 X.describe(include="all")
 
@@ -64,6 +65,8 @@ X.head()
 # %%
 # Our target for prediction: the wage.
 # Wages are described as floating-point number in dollars per hour.
+
+# %%
 y = survey.target.values.ravel()
 survey.target.head()
 
@@ -121,7 +124,7 @@ survey.data.info()
 # Our pre-processor will
 #
 # - one-hot encode (i.e., generate a column by category) the categorical
-#   columns;
+#   columns, only for non-binary categorical variables;
 # - as a first approach (we will see after how the normalisation of numerical
 #   values will affect our discussion), keep numerical values as they are.
 
@@ -134,14 +137,13 @@ numerical_columns = ["EDUCATION", "EXPERIENCE", "AGE"]
 preprocessor = make_column_transformer(
     (OneHotEncoder(drop="if_binary"), categorical_columns),
     remainder="passthrough",
-    verbose_feature_names_out=False,
+    verbose_feature_names_out=False,  # avoid to prepend the preprocessor names
 )
 
 # %%
 # To describe the dataset as a linear model we use a ridge regressor
 # with a very small regularization and to model the logarithm of the WAGE.
 
-
 from sklearn.pipeline import make_pipeline
 from sklearn.linear_model import Ridge
 from sklearn.compose import TransformedTargetRegressor
@@ -159,7 +161,7 @@ model = make_pipeline(
 #
 # First, we fit the model.
 
-_ = model.fit(X_train, y_train)
+model.fit(X_train, y_train)
 
 # %%
 # Then we check the performance of the computed model plotting its predictions
@@ -167,28 +169,31 @@ _ = model.fit(X_train, y_train)
 # for example, the median absolute error of the model.
 
 from sklearn.metrics import median_absolute_error
+from sklearn.metrics import PredictionErrorDisplay
 
-y_pred = model.predict(X_train)
-
-mae = median_absolute_error(y_train, y_pred)
-string_score = f"MAE on training set: {mae:.2f} $/hour"
+mae_train = median_absolute_error(y_train, model.predict(X_train))
 y_pred = model.predict(X_test)
-mae = median_absolute_error(y_test, y_pred)
-string_score += f"\nMAE on testing set: {mae:.2f} $/hour"
-fig, ax = plt.subplots(figsize=(5, 5))
-plt.scatter(y_test, y_pred)
-ax.plot([0, 1], [0, 1], transform=ax.transAxes, ls="--", c="red")
-plt.text(3, 20, string_score)
-plt.title("Ridge model, small regularization")
-plt.ylabel("Model predictions")
-plt.xlabel("Truths")
-plt.xlim([0, 27])
-_ = plt.ylim([0, 27])
+mae_test = median_absolute_error(y_test, y_pred)
+scores = {
+    "MedAE on training set": f"{mae_train:.2f} $/hour",
+    "MedAE on testing set": f"{mae_test:.2f} $/hour",
+}
+
+# %%
+_, ax = plt.subplots(figsize=(5, 5))
+display = PredictionErrorDisplay.from_predictions(
+    y_test, y_pred, kind="actual_vs_predicted", ax=ax, scatter_kwargs={"alpha": 0.5}
+)
+ax.set_title("Ridge model, small regularization")
+for name, score in scores.items():
+    ax.plot([], [], " ", label=f"{name}: {score}")
+ax.legend(loc="upper left")
+plt.tight_layout()
 
 # %%
 # The model learnt is far from being a good model making accurate predictions:
 # this is obvious when looking at the plot above, where good predictions
-# should lie on the red line.
+# should lie on the black dashed line.
 #
 # In the following section, we will interpret the coefficients of the model.
 # While we do so, we should keep in mind that any conclusion we draw is
@@ -196,14 +201,14 @@ _ = plt.ylim([0, 27])
 # generative process of the data.
 #
 # Interpreting coefficients: scale matters
-# ---------------------------------------------
+# ----------------------------------------
 #
 # First of all, we can take a look to the values of the coefficients of the
 # regressor we have fitted.
 feature_names = model[:-1].get_feature_names_out()
 
 coefs = pd.DataFrame(
-    model.named_steps["transformedtargetregressor"].regressor_.coef_,
+    model[-1].regressor_.coef_,
     columns=["Coefficients"],
     index=feature_names,
 )
@@ -224,9 +229,10 @@ coefs
 # hence value ranges, because of their different unit of measure. This is more
 # visible if we plot the coefficients.
 
-coefs.plot(kind="barh", figsize=(9, 7))
+coefs.plot.barh(figsize=(9, 7))
 plt.title("Ridge model, small regularization")
 plt.axvline(x=0, color=".5")
+plt.xlabel("Raw coefficient values")
 plt.subplots_adjust(left=0.3)
 
 # %%
@@ -243,11 +249,12 @@ plt.subplots_adjust(left=0.3)
 # features.
 
 X_train_preprocessed = pd.DataFrame(
-    model.named_steps["columntransformer"].transform(X_train), columns=feature_names
+    model[:-1].transform(X_train), columns=feature_names
 )
 
-X_train_preprocessed.std(axis=0).plot(kind="barh", figsize=(9, 7))
-plt.title("Features std. dev.")
+X_train_preprocessed.std(axis=0).plot.barh(figsize=(9, 7))
+plt.title("Feature ranges")
+plt.xlabel("Std. dev. of feature values")
 plt.subplots_adjust(left=0.3)
 
 # %%
@@ -263,12 +270,12 @@ plt.subplots_adjust(left=0.3)
 # coefficient on the output, all else being equal.
 
 coefs = pd.DataFrame(
-    model.named_steps["transformedtargetregressor"].regressor_.coef_
-    * X_train_preprocessed.std(axis=0),
+    model[-1].regressor_.coef_ * X_train_preprocessed.std(axis=0),
     columns=["Coefficient importance"],
     index=feature_names,
 )
 coefs.plot(kind="barh", figsize=(9, 7))
+plt.xlabel("Coefficient values corrected by the feature's std. dev.")
 plt.title("Ridge model, small regularization")
 plt.axvline(x=0, color=".5")
 plt.subplots_adjust(left=0.3)
@@ -305,28 +312,32 @@ plt.subplots_adjust(left=0.3)
 from sklearn.model_selection import cross_validate
 from sklearn.model_selection import RepeatedKFold
 
+cv = RepeatedKFold(n_splits=5, n_repeats=5, random_state=0)
 cv_model = cross_validate(
     model,
     X,
     y,
-    cv=RepeatedKFold(n_splits=5, n_repeats=5),
+    cv=cv,
     return_estimator=True,
     n_jobs=2,
 )
+
 coefs = pd.DataFrame(
     [
-        est.named_steps["transformedtargetregressor"].regressor_.coef_
-        * X_train_preprocessed.std(axis=0)
-        for est in cv_model["estimator"]
+        est[-1].regressor_.coef_ * est[:-1].transform(X.iloc[train_idx]).std(axis=0)
+        for est, (train_idx, _) in zip(cv_model["estimator"], cv.split(X, y))
     ],
     columns=feature_names,
 )
+
+# %%
 plt.figure(figsize=(9, 7))
-sns.stripplot(data=coefs, orient="h", color="k", alpha=0.5)
-sns.boxplot(data=coefs, orient="h", color="cyan", saturation=0.5)
+sns.stripplot(data=coefs, orient="h", palette="dark:k", alpha=0.5)
+sns.boxplot(data=coefs, orient="h", color="cyan", saturation=0.5, whis=10)
 plt.axvline(x=0, color=".5")
 plt.xlabel("Coefficient importance")
 plt.title("Coefficient importance and its variability")
+plt.suptitle("Ridge model, small regularization")
 plt.subplots_adjust(left=0.3)
 
 # %%
@@ -364,29 +375,34 @@ cv_model = cross_validate(
     model,
     X.drop(columns=column_to_drop),
     y,
-    cv=RepeatedKFold(n_splits=5, n_repeats=5),
+    cv=cv,
     return_estimator=True,
     n_jobs=2,
 )
+
 coefs = pd.DataFrame(
     [
-        est.named_steps["transformedtargetregressor"].regressor_.coef_
-        * X_train_preprocessed.drop(columns=column_to_drop).std(axis=0)
-        for est in cv_model["estimator"]
+        est[-1].regressor_.coef_
+        * est[:-1].transform(X.drop(columns=column_to_drop).iloc[train_idx]).std(axis=0)
+        for est, (train_idx, _) in zip(cv_model["estimator"], cv.split(X, y))
     ],
     columns=feature_names[:-1],
 )
+
+# %%
 plt.figure(figsize=(9, 7))
-sns.stripplot(data=coefs, orient="h", color="k", alpha=0.5)
+sns.stripplot(data=coefs, orient="h", palette="dark:k", alpha=0.5)
 sns.boxplot(data=coefs, orient="h", color="cyan", saturation=0.5)
 plt.axvline(x=0, color=".5")
 plt.title("Coefficient importance and its variability")
 plt.xlabel("Coefficient importance")
+plt.suptitle("Ridge model, small regularization, AGE dropped")
 plt.subplots_adjust(left=0.3)
 
 # %%
-# The estimation of the EXPERIENCE coefficient is now less variable and
-# remain important for all models trained during cross-validation.
+# The estimation of the EXPERIENCE coefficient now shows a much reduced
+# variability. EXPERIENCE remains important for all models trained during
+# cross-validation.
 #
 # .. _scaling_num:
 #
@@ -395,8 +411,8 @@ plt.subplots_adjust(left=0.3)
 #
 # As said above (see ":ref:`the-pipeline`"), we could also choose to scale
 # numerical values before training the model.
-# This can be useful to apply a similar amount regularization to all of them
-# in the Ridge.
+# This can be useful when we apply a similar amount of regularization to all of them
+# in the ridge.
 # The preprocessor is redefined in order to subtract the mean and scale
 # variables to unit variance.
 
@@ -405,7 +421,6 @@ from sklearn.preprocessing import StandardScaler
 preprocessor = make_column_transformer(
     (OneHotEncoder(drop="if_binary"), categorical_columns),
     (StandardScaler(), numerical_columns),
-    remainder="passthrough",
 )
 
 # %%
@@ -417,66 +432,68 @@ model = make_pipeline(
         regressor=Ridge(alpha=1e-10), func=np.log10, inverse_func=sp.special.exp10
     ),
 )
-
-_ = model.fit(X_train, y_train)
+model.fit(X_train, y_train)
 
 # %%
 # Again, we check the performance of the computed
 # model using, for example, the median absolute error of the model and the R
 # squared coefficient.
 
-y_pred = model.predict(X_train)
-mae = median_absolute_error(y_train, y_pred)
-string_score = f"MAE on training set: {mae:.2f} $/hour"
+mae_train = median_absolute_error(y_train, model.predict(X_train))
 y_pred = model.predict(X_test)
-mae = median_absolute_error(y_test, y_pred)
-string_score += f"\nMAE on testing set: {mae:.2f} $/hour"
-fig, ax = plt.subplots(figsize=(6, 6))
-plt.scatter(y_test, y_pred)
-ax.plot([0, 1], [0, 1], transform=ax.transAxes, ls="--", c="red")
+mae_test = median_absolute_error(y_test, y_pred)
+scores = {
+    "MedAE on training set": f"{mae_train:.2f} $/hour",
+    "MedAE on testing set": f"{mae_test:.2f} $/hour",
+}
 
-plt.text(3, 20, string_score)
-
-plt.title("Ridge model, small regularization, normalized variables")
-plt.ylabel("Model predictions")
-plt.xlabel("Truths")
-plt.xlim([0, 27])
-_ = plt.ylim([0, 27])
+_, ax = plt.subplots(figsize=(5, 5))
+display = PredictionErrorDisplay.from_predictions(
+    y_test, y_pred, kind="actual_vs_predicted", ax=ax, scatter_kwargs={"alpha": 0.5}
+)
+ax.set_title("Ridge model, small regularization")
+for name, score in scores.items():
+    ax.plot([], [], " ", label=f"{name}: {score}")
+ax.legend(loc="upper left")
+plt.tight_layout()
 
 # %%
-# For the coefficient analysis, scaling is not needed this time.
+# For the coefficient analysis, scaling is not needed this time because it
+# was performed during the preprocessing step.
 
 coefs = pd.DataFrame(
-    model.named_steps["transformedtargetregressor"].regressor_.coef_,
-    columns=["Coefficients"],
+    model[-1].regressor_.coef_,
+    columns=["Coefficients importance"],
     index=feature_names,
 )
-coefs.plot(kind="barh", figsize=(9, 7))
+coefs.plot.barh(figsize=(9, 7))
 plt.title("Ridge model, small regularization, normalized variables")
+plt.xlabel("Raw coefficient values")
 plt.axvline(x=0, color=".5")
 plt.subplots_adjust(left=0.3)
 
 # %%
-# We now inspect the coefficients across several cross-validation folds.
+# We now inspect the coefficients across several cross-validation folds. As in
+# the above example, we do not need to scale the coefficients by the std. dev.
+# of the feature values since this scaling was already
+# done in the preprocessing step of the pipeline.
 
 cv_model = cross_validate(
     model,
     X,
     y,
-    cv=RepeatedKFold(n_splits=5, n_repeats=5),
+    cv=cv,
     return_estimator=True,
     n_jobs=2,
 )
 coefs = pd.DataFrame(
-    [
-        est.named_steps["transformedtargetregressor"].regressor_.coef_
-        for est in cv_model["estimator"]
-    ],
-    columns=feature_names,
+    [est[-1].regressor_.coef_ for est in cv_model["estimator"]], columns=feature_names
 )
+
+# %%
 plt.figure(figsize=(9, 7))
-sns.stripplot(data=coefs, orient="h", color="k", alpha=0.5)
-sns.boxplot(data=coefs, orient="h", color="cyan", saturation=0.5)
+sns.stripplot(data=coefs, orient="h", palette="dark:k", alpha=0.5)
+sns.boxplot(data=coefs, orient="h", color="cyan", saturation=0.5, whis=10)
 plt.axvline(x=0, color=".5")
 plt.title("Coefficient variability")
 plt.subplots_adjust(left=0.3)
@@ -487,27 +504,27 @@ plt.subplots_adjust(left=0.3)
 # Linear models with regularization
 # ---------------------------------
 #
-# In machine-learning practice, Ridge Regression is more often used with
+# In machine-learning practice, ridge regression is more often used with
 # non-negligible regularization.
 #
-# Above, we limited this regularization to a very little amount.
-# Regularization improves the conditioning of the problem and reduces the
-# variance of the estimates. RidgeCV applies cross validation in order to
-# determine which value of the regularization parameter (`alpha`) is best
-# suited for prediction.
+# Above, we limited this regularization to a very little amount. Regularization
+# improves the conditioning of the problem and reduces the variance of the
+# estimates. :class:`~sklearn.linear_model.RidgeCV` applies cross validation
+# in order to determine which value of the regularization parameter (`alpha`)
+# is best suited for prediction.
 
 from sklearn.linear_model import RidgeCV
 
+alphas = np.logspace(-10, 10, 21)  # alpha values to be chosen from by cross-validation
 model = make_pipeline(
     preprocessor,
     TransformedTargetRegressor(
-        regressor=RidgeCV(alphas=np.logspace(-10, 10, 21)),
+        regressor=RidgeCV(alphas=alphas),
         func=np.log10,
         inverse_func=sp.special.exp10,
     ),
 )
-
-_ = model.fit(X_train, y_train)
+model.fit(X_train, y_train)
 
 # %%
 # First we check which value of :math:`\alpha` has been selected.
@@ -516,37 +533,36 @@ model[-1].regressor_.alpha_
 
 # %%
 # Then we check the quality of the predictions.
-
-y_pred = model.predict(X_train)
-mae = median_absolute_error(y_train, y_pred)
-string_score = f"MAE on training set: {mae:.2f} $/hour"
+mae_train = median_absolute_error(y_train, model.predict(X_train))
 y_pred = model.predict(X_test)
-mae = median_absolute_error(y_test, y_pred)
-string_score += f"\nMAE on testing set: {mae:.2f} $/hour"
+mae_test = median_absolute_error(y_test, y_pred)
+scores = {
+    "MedAE on training set": f"{mae_train:.2f} $/hour",
+    "MedAE on testing set": f"{mae_test:.2f} $/hour",
+}
 
-fig, ax = plt.subplots(figsize=(6, 6))
-plt.scatter(y_test, y_pred)
-ax.plot([0, 1], [0, 1], transform=ax.transAxes, ls="--", c="red")
-
-plt.text(3, 20, string_score)
-
-plt.title("Ridge model, regularization, normalized variables")
-plt.ylabel("Model predictions")
-plt.xlabel("Truths")
-plt.xlim([0, 27])
-_ = plt.ylim([0, 27])
+_, ax = plt.subplots(figsize=(5, 5))
+display = PredictionErrorDisplay.from_predictions(
+    y_test, y_pred, kind="actual_vs_predicted", ax=ax, scatter_kwargs={"alpha": 0.5}
+)
+ax.set_title("Ridge model, optimum regularization")
+for name, score in scores.items():
+    ax.plot([], [], " ", label=f"{name}: {score}")
+ax.legend(loc="upper left")
+plt.tight_layout()
 
 # %%
 # The ability to reproduce the data of the regularized model is similar to
 # the one of the non-regularized model.
 
 coefs = pd.DataFrame(
-    model.named_steps["transformedtargetregressor"].regressor_.coef_,
-    columns=["Coefficients"],
+    model[-1].regressor_.coef_,
+    columns=["Coefficients importance"],
     index=feature_names,
 )
-coefs.plot(kind="barh", figsize=(9, 7))
-plt.title("Ridge model, regularization, normalized variables")
+coefs.plot.barh(figsize=(9, 7))
+plt.title("Ridge model, with regularization, normalized variables")
+plt.xlabel("Raw coefficient values")
 plt.axvline(x=0, color=".5")
 plt.subplots_adjust(left=0.3)
 
@@ -560,28 +576,24 @@ plt.subplots_adjust(left=0.3)
 # predictive variables, so neither alone would have strong weights.
 #
 # On the other hand, the weights obtained with regularization are more
-# stable  (see the :ref:`ridge_regression` User Guide section). This
+# stable (see the :ref:`ridge_regression` User Guide section). This
 # increased stability is visible from the plot, obtained from data
-# perturbations, in a cross validation. This plot can  be compared with
+# perturbations, in a cross-validation. This plot can be compared with
 # the :ref:`previous one<covariation>`.
 
 cv_model = cross_validate(
     model,
     X,
     y,
-    cv=RepeatedKFold(n_splits=5, n_repeats=5),
+    cv=cv,
     return_estimator=True,
     n_jobs=2,
 )
 coefs = pd.DataFrame(
-    [
-        est.named_steps["transformedtargetregressor"].regressor_.coef_
-        * X_train_preprocessed.std(axis=0)
-        for est in cv_model["estimator"]
-    ],
-    columns=feature_names,
+    [est[-1].regressor_.coef_ for est in cv_model["estimator"]], columns=feature_names
 )
 
+# %%
 plt.ylabel("Age coefficient")
 plt.xlabel("Experience coefficient")
 plt.grid(True)
@@ -596,19 +608,20 @@ _ = plt.title("Co-variations of coefficients for AGE and EXPERIENCE across folds
 #
 # Another possibility to take into account correlated variables in the dataset,
 # is to estimate sparse coefficients. In some way we already did it manually
-# when we dropped the AGE column in a previous Ridge estimation.
+# when we dropped the AGE column in a previous ridge estimation.
 #
 # Lasso models (see the :ref:`lasso` User Guide section) estimates sparse
-# coefficients. LassoCV applies cross validation in order to
-# determine which value of the regularization parameter (`alpha`) is best
-# suited for the model estimation.
+# coefficients. :class:`~sklearn.linear_model.LassoCV` applies cross
+# validation in order to determine which value of the regularization parameter
+# (`alpha`) is best suited for the model estimation.
 
 from sklearn.linear_model import LassoCV
 
+alphas = np.logspace(-10, 10, 21)  # alpha values to be chosen from by cross-validation
 model = make_pipeline(
     preprocessor,
     TransformedTargetRegressor(
-        regressor=LassoCV(alphas=np.logspace(-10, 10, 21), max_iter=100000),
+        regressor=LassoCV(alphas=alphas, max_iter=100_000),
         func=np.log10,
         inverse_func=sp.special.exp10,
     ),
@@ -624,35 +637,34 @@ model[-1].regressor_.alpha_
 # %%
 # Then we check the quality of the predictions.
 
-y_pred = model.predict(X_train)
-mae = median_absolute_error(y_train, y_pred)
-string_score = f"MAE on training set: {mae:.2f} $/hour"
+mae_train = median_absolute_error(y_train, model.predict(X_train))
 y_pred = model.predict(X_test)
-mae = median_absolute_error(y_test, y_pred)
-string_score += f"\nMAE on testing set: {mae:.2f} $/hour"
+mae_test = median_absolute_error(y_test, y_pred)
+scores = {
+    "MedAE on training set": f"{mae_train:.2f} $/hour",
+    "MedAE on testing set": f"{mae_test:.2f} $/hour",
+}
 
-fig, ax = plt.subplots(figsize=(6, 6))
-plt.scatter(y_test, y_pred)
-ax.plot([0, 1], [0, 1], transform=ax.transAxes, ls="--", c="red")
-
-plt.text(3, 20, string_score)
-
-plt.title("Lasso model, regularization, normalized variables")
-plt.ylabel("Model predictions")
-plt.xlabel("Truths")
-plt.xlim([0, 27])
-_ = plt.ylim([0, 27])
+_, ax = plt.subplots(figsize=(6, 6))
+display = PredictionErrorDisplay.from_predictions(
+    y_test, y_pred, kind="actual_vs_predicted", ax=ax, scatter_kwargs={"alpha": 0.5}
+)
+ax.set_title("Lasso model, optimum regularization")
+for name, score in scores.items():
+    ax.plot([], [], " ", label=f"{name}: {score}")
+ax.legend(loc="upper left")
+plt.tight_layout()
 
 # %%
 # For our dataset, again the model is not very predictive.
 
 coefs = pd.DataFrame(
-    model.named_steps["transformedtargetregressor"].regressor_.coef_,
-    columns=["Coefficients"],
+    model[-1].regressor_.coef_,
+    columns=["Coefficients importance"],
     index=feature_names,
 )
 coefs.plot(kind="barh", figsize=(9, 7))
-plt.title("Lasso model, regularization, normalized variables")
+plt.title("Lasso model, optimum regularization, normalized variables")
 plt.axvline(x=0, color=".5")
 plt.subplots_adjust(left=0.3)
 
@@ -667,6 +679,31 @@ plt.subplots_adjust(left=0.3)
 # is unstable for correlated features, and should be interpreted with
 # caution.
 #
+# Indeed, we can check the variability of the coefficients across folds.
+cv_model = cross_validate(
+    model,
+    X,
+    y,
+    cv=cv,
+    return_estimator=True,
+    n_jobs=2,
+)
+coefs = pd.DataFrame(
+    [est[-1].regressor_.coef_ for est in cv_model["estimator"]], columns=feature_names
+)
+
+# %%
+plt.figure(figsize=(9, 7))
+sns.stripplot(data=coefs, orient="h", palette="dark:k", alpha=0.5)
+sns.boxplot(data=coefs, orient="h", color="cyan", saturation=0.5, whis=100)
+plt.axvline(x=0, color=".5")
+plt.title("Coefficient variability")
+plt.subplots_adjust(left=0.3)
+
+# %%
+# We observe that the AGE and EXPERIENCE coefficients are varying a lot
+# depending of the fold.
+#
 # Lessons learned
 # ---------------
 #
